@@ -6,9 +6,24 @@
 #include "DynamsoftCommon.h"
 #include "DynamsoftBarcodeReader.h"
 #include "barcode_result.h"
+#include <thread>
+#include <condition_variable>
+#include <mutex>
+#include <queue>
+#include <functional>
 
 #define DEBUG 0
 
+class WorkerThread
+{
+public:
+    std::mutex m;
+    std::condition_variable cv;
+    std::queue<std::function<void()>> tasks = {};
+    volatile bool running;
+	std::thread t;
+};
+
 typedef struct
 {
     PyObject_HEAD
@@ -16,15 +31,55 @@ typedef struct
     void *hBarcode;
     // Callback function for video mode
     PyObject *py_cb_textResult;
+    // Callback function for image mode
+    PyObject *callback;
+    // Worker thread
+    WorkerThread *worker;
 } DynamsoftBarcodeReader;
 
+void clearTasks(DynamsoftBarcodeReader *self)
+{
+    if (self->callback)
+    {
+        Py_XDECREF(self->callback);
+        self->callback = NULL;
+    }
+
+    if (self->py_cb_textResult) 
+    {
+        Py_XDECREF(self->py_cb_textResult);
+        self->py_cb_textResult = NULL;
+    }
+
+    if (self->worker)
+    {
+        std::unique_lock<std::mutex> lk(self->worker->m);
+        self->worker->running = false;
+        
+        if (self->worker->tasks.size() > 1)
+        {
+            std::queue<std::function<void()>> empty = {};
+            std::swap(self->worker->tasks, empty);
+        }
+
+        self->worker->cv.notify_one();
+        lk.unlock();
+
+        self->worker->t.join();
+        delete self->worker;
+        self->worker = NULL;
+        printf("Quit native thread.\n");
+    }
+}
+
 static int DynamsoftBarcodeReader_clear(DynamsoftBarcodeReader *self)
 {
-    if (self->py_cb_textResult) Py_XDECREF(self->py_cb_textResult);
+    clearTasks(self);
     if(self->hBarcode) {
 		DBR_DestroyInstance(self->hBarcode);
     	self->hBarcode = NULL;
 	}
+
     return 0;
 }
 
@@ -43,6 +98,8 @@ static PyObject *DynamsoftBarcodeReader_new(PyTypeObject *type, PyObject *args,
     {
        	self->hBarcode = DBR_CreateInstance();
         self->py_cb_textResult = NULL;
+        self->callback = NULL;
+        self->worker = NULL;
     }
 
     return (PyObject *)self;
@@ -202,6 +259,237 @@ static PyObject *decodeMat(PyObject *obj, PyObject *args)
     return list;
 }
 
+void onResultReady(DynamsoftBarcodeReader *self)
+{
+    PyGILState_STATE gstate;
+    gstate = PyGILState_Ensure();
+    PyObject *list = createPyResults(self);
+    PyObject *result = PyObject_CallFunction(self->callback, "O", list);
+    if (result != NULL)
+        Py_DECREF(result);
+
+    PyGILState_Release(gstate);
+}
+
+void scan(DynamsoftBarcodeReader *self, unsigned char *buffer, int width, int height, int stride, ImagePixelFormat format, int len)
+{
+    ImageData data;
+    data.bytes = buffer;
+    data.width = width;
+    data.height = height;
+    data.stride = stride;
+    data.format = format;
+    data.bytesLength = len;
+
+    int ret = DBR_DecodeBuffer(self->hBarcode, (const unsigned char*)buffer, width, height, stride, format, "");
+    if (ret)
+    {
+        printf("Detection error: %s\n", DBR_GetErrorString(ret));
+    }
+
+    free(buffer);
+
+    if (self->callback != NULL)
+    {
+        onResultReady(self);
+    }
+}
+
+/**
+ * Decode barcode and QR code from OpenCV Mat asynchronously.
+ *
+ * @param Mat image
+ *
+ */
+static PyObject *decodeMatAsync(PyObject *obj, PyObject *args)
+{
+    DynamsoftBarcodeReader *self = (DynamsoftBarcodeReader *)obj;
+    if (self->worker == NULL)
+    {
+        return Py_BuildValue("i", -1);
+    }
+    PyObject *o;
+    if (!PyArg_ParseTuple(args, "O", &o))
+        return Py_BuildValue("i", -1);
+
+    Py_buffer *view;
+    int nd;
+    PyObject *memoryview = PyMemoryView_FromObject(o);
+    if (memoryview == NULL)
+    {
+        PyErr_Clear();
+        return Py_BuildValue("i", -1);
+    }
+
+    view = PyMemoryView_GET_BUFFER(memoryview);
+    char *buffer = (char *)view->buf;
+    nd = view->ndim;
+    int len = view->len;
+    int stride = view->strides[0];
+    int width = view->strides[0] / view->strides[1];
+    int height = len / stride;
+
+    ImagePixelFormat format = IPF_RGB_888;
+
+    if (width == stride)
+    {
+        format = IPF_GRAYSCALED;
+    }
+    else if (width * 3 == stride)
+    {
+        format = IPF_RGB_888;
+    }
+    else if (width * 4 == stride)
+    {
+        format = IPF_ARGB_8888;
+    }
+
+    unsigned char *data = (unsigned char *)malloc(len);
+    memcpy(data, buffer, len);
+
+    std::unique_lock<std::mutex> lk(self->worker->m);
+    if (self->worker->tasks.size() > 1)
+    {
+        std::queue<std::function<void()>> empty = {};
+        std::swap(self->worker->tasks, empty);
+    }
+    std::function<void()> task_function = std::bind(scan, self, data, width, height, stride, format, len);
+    self->worker->tasks.push(task_function);
+    self->worker->cv.notify_one();
+    lk.unlock();
+    
+    Py_DECREF(memoryview);
+    return Py_BuildValue("i", 0);
+}
+
+/**
+ * Decode barcode and QR code from byte array asynchronously.
+ *
+ * @param bytes, width, height, stride, imageformat
+ *
+ */
+static PyObject *decodeBytesAsync(PyObject * obj, PyObject *args)
+{
+    DynamsoftBarcodeReader *self = (DynamsoftBarcodeReader *)obj;  
+    if (self->worker == NULL)
+    {
+        return Py_BuildValue("i", -1);
+    }
+
+    PyObject *o;
+    int len, width, height, stride, imageformat;
+    if (!PyArg_ParseTuple(args, "Oiiiii", &o, &len, &width, &height, &stride, &imageformat))
+		return Py_BuildValue("i", -1);
+
+    char * barcodeBuffer = NULL;
+    if(PyByteArray_Check(o))
+    {
+        barcodeBuffer = PyByteArray_AsString(o);
+    }
+    else if(PyBytes_Check(o))
+    {
+        barcodeBuffer = PyBytes_AsString(o);
+    }
+    else
+    {
+        printf("The first parameter should be a byte array or bytes object.");
+		return Py_BuildValue("i", -1);
+    }
+    
+    ImagePixelFormat format = IPF_RGB_888;
+
+    if (imageformat == 0)
+    {
+        format = IPF_GRAYSCALED;
+    }
+    else
+    {
+        format = IPF_RGB_888;
+    }
+    
+    unsigned char *data = (unsigned char *)malloc(len);
+    memcpy(data, barcodeBuffer, len);
+
+    std::unique_lock<std::mutex> lk(self->worker->m);
+    if (self->worker->tasks.size() > 1)
+    {
+        std::queue<std::function<void()>> empty = {};
+        std::swap(self->worker->tasks, empty);
+    }
+    std::function<void()> task_function = std::bind(scan, self, data, width, height, stride, format, len);
+    self->worker->tasks.push(task_function);
+    self->worker->cv.notify_one();
+    lk.unlock();
+
+    return Py_BuildValue("i", 0);
+}
+
+void run(DynamsoftBarcodeReader *self)
+{
+    while (self->worker->running)
+    {
+        std::function<void()> task;
+        std::unique_lock<std::mutex> lk(self->worker->m);
+        self->worker->cv.wait(lk, [&]
+                              { return !self->worker->tasks.empty() || !self->worker->running; });
+        if (!self->worker->running)
+		{
+			break;
+		}
+        task = std::move(self->worker->tasks.front());
+        self->worker->tasks.pop();
+        lk.unlock();
+
+        task();
+    }
+}
+
+/**
+ * Register callback function to receive barcode decoding result asynchronously.
+ */
+static PyObject *addAsyncListener(PyObject *obj, PyObject *args)
+{
+    DynamsoftBarcodeReader *self = (DynamsoftBarcodeReader *)obj;
+
+    PyObject *callback = NULL;
+    if (!PyArg_ParseTuple(args, "O", &callback))
+    {
+        return NULL;
+    }
+
+    if (!PyCallable_Check(callback))
+    {
+        PyErr_SetString(PyExc_TypeError, "parameter must be callable");
+        return NULL;
+    }
+    else
+    {
+        Py_XINCREF(callback);       /* Add a reference to new callback */
+        Py_XDECREF(self->callback); /* Dispose of previous callback */
+        self->callback = callback;
+    }
+
+    if (self->worker == NULL)
+    {
+        self->worker = new WorkerThread();
+        self->worker->running = true;
+        self->worker->t = std::thread(&run, self);
+    }
+
+    printf("Running native thread...\n");
+    return Py_BuildValue("i", 0);
+}
+
+/**
+ * Clear native thread and tasks.
+ */
+static PyObject *clearAsyncListener(PyObject *obj, PyObject *args)
+{
+    DynamsoftBarcodeReader *self = (DynamsoftBarcodeReader *)obj;
+    clearTasks(self);
+    return Py_BuildValue("i", 0);
+}
+
 /**
  * Get runtime settings.
  *
@@ -252,9 +540,8 @@ static PyObject *decodeBytes(PyObject * obj, PyObject *args)
     DynamsoftBarcodeReader *self = (DynamsoftBarcodeReader *)obj;  
 
     PyObject *o;
-    int width, height, stride;
-    ImagePixelFormat imagePixelFormat = IPF_RGB_888;
-    if (!PyArg_ParseTuple(args, "Oiiii", &o, &height, &width, &stride, &imagePixelFormat))
+    int width, height, stride, imageformat;
+    if (!PyArg_ParseTuple(args, "Oiiii", &o, &width, &height, &stride, &imageformat))
 		Py_RETURN_NONE;
 
     char * barcodeBuffer = NULL;
@@ -272,7 +559,18 @@ static PyObject *decodeBytes(PyObject * obj, PyObject *args)
 		Py_RETURN_NONE;
     }
     
-    int ret = DBR_DecodeBuffer(self->hBarcode, (const unsigned char*)barcodeBuffer, width, height, stride, imagePixelFormat, "");
+    ImagePixelFormat format = IPF_RGB_888;
+
+    if (imageformat == 0)
+    {
+        format = IPF_GRAYSCALED;
+    }
+    else
+    {
+        format = IPF_RGB_888;
+    }
+    
+    int ret = DBR_DecodeBuffer(self->hBarcode, (const unsigned char*)barcodeBuffer, width, height, stride, format, "");
     PyObject *list = createPyResults(self);
 
     return list;
@@ -393,6 +691,10 @@ static PyMethodDef instance_methods[] = {
   {"stopVideoMode", stopVideoMode, METH_VARARGS, NULL},
   {"appendVideoFrame", appendVideoFrame, METH_VARARGS, NULL},
   {"decodeBytes", decodeBytes, METH_VARARGS, NULL},
+  {"addAsyncListener", addAsyncListener, METH_VARARGS, NULL},
+  {"decodeMatAsync", decodeMatAsync, METH_VARARGS, NULL},
+  {"clearAsyncListener", clearAsyncListener, METH_VARARGS, NULL},
+  {"decodeBytesAsync", decodeBytesAsync, METH_VARARGS, NULL},
   {NULL, NULL, 0, NULL}       
 };
 
